import json
import random
import string
from typing import Optional
from urllib.parse import urljoin

import requests
from cryptography.utils import cached_property
from pydantic import BaseModel, Field, PrivateAttr

from pysilpo.utils.cursor import Cursor
from pysilpo.utils.exceptions import SilpoRequestException

_GRAPHQL_API_URL = "https://graphql.silpo.ua/graphql"


class CityModel(BaseModel):
    _stores: Cursor["StoreModel"] = PrivateAttr(None)
    id: str
    title: str
    slug: str

    def __init__(self, **data):
        super().__init__(**data)
        stores = data.pop("storeFilterable", None)
        if stores is not None:
            stores = [StoreModel(**x) for x in stores]
            self._stores = Cursor(generator=lambda _offset: (stores, len(stores)), page_size=len(stores))

    @cached_property
    def stores(self) -> Cursor["StoreModel"]:
        if not self._stores:
            self._stores = Store.all(city_id=self.id)
        return self._stores


class StoreModel(BaseModel):
    def __init__(self, **data):
        super().__init__(**data)
        filial_id = data.get("filial_id")
        if filial_id is not None:
            self.filial_id = filial_id

    id: str = Field(..., alias="id")
    images: list[dict] = Field(..., alias="images")
    electricity_state: int = Field(..., alias="electricityState")
    is_designed: bool = Field(..., alias="isDesigned")
    is_lesilpo: bool = Field(..., alias="isLesilpo")
    filial_id: Optional[int] = Field(None, alias="filialId")
    link: Optional[str] = Field(None, alias="link")
    title: str = Field(..., alias="title")
    premium: bool = Field(..., alias="premium")
    map_link: Optional[str] = Field(None, alias="mapLink")
    slug: str = Field(..., alias="slug")
    active: bool = Field(..., alias="active")
    cache_amount: int = Field(..., alias="cacheAmount")
    terminal_enabled: bool = Field(..., alias="terminalEnabled")
    with_generator: bool = Field(..., alias="withGenerator")
    with_wifi: bool = Field(..., alias="withWifi")
    with_starlink: bool = Field(..., alias="withStarlink")
    active_hours: dict = Field(..., alias="activeHours")
    filial_type: str = Field(..., alias="filialType")
    location: dict = Field(..., alias="location")
    city: CityModel
    updated_at: Optional[str] = Field(None, alias="updatedAt")

    @cached_property
    def branch_id(self) -> Optional[str]:
        try:
            if self.filial_id is None:
                return None
            return Store.get_branch_id(self.filial_id)[0].branch_id
        except IndexError:
            raise SilpoRequestException(f"Branch not found for filial_id: {self.filial_id}") from None


class FilialModel(BaseModel):
    branch_id: str = Field(..., alias="branchId")
    company_id: str = Field(..., alias="companyId")
    filial_id: str = Field(..., alias="filialId")


class Store:
    _BASE_RESTFUL_DOMAIN = "https://sf-ecom-api.silpo.ua"

    _GET_BRANCH_BY_FILIAL_ID_URL = urljoin(_BASE_RESTFUL_DOMAIN, "/v1/branches/by-filial-ids")

    _ALL_STORES_QUERY = """query stores($filter: StoreFilterInputType, $pagingInfo: InputBatch!) {
      stores(filter: $filter, pagingInfo: $pagingInfo) {
        limit
        offset
        count
        items {
          ...StoreBaseFragment
          updatedAt
          __typename
        }
        __typename
      }
    }

    fragment StoreBaseFragment on Store {
      id
      images {
        image {
          url
          __typename
        }
        __typename
      }
      electricityState
      isDesigned
      isLesilpo
      filial_id
      link
      title
      premium
      mapLink
      slug
      active
      cacheAmount
      terminalEnabled
      withGenerator
      withWifi
      withStarlink
      activeHours {
        start
        end
        __typename
      }
      filialType
      location {
        ...LocationBaseFragment
        __typename
      }
      city {
        ...CityBaseFragment
        __typename
      }
      __typename
    }

    fragment CityBaseFragment on City {
      id
      title
      slug
      __typename
    }

    fragment LocationBaseFragment on Location {
      lat
      lng
      __typename
    }"""

    @classmethod
    def all(cls, city_id: Optional[str] = None) -> Cursor[StoreModel]:
        form_data = {
            "query": cls._ALL_STORES_QUERY,
            "variables": {
                "filter": {
                    "filialId": None,
                    "cityId": city_id,
                    "start": None,
                    "end": None,
                    "hasCertificate": None,
                    "servicesIds": None,
                },
                "pagingInfo": {"limit": 400, "offset": 0},
            },
            "operationName": "stores",
        }

        # Custom boundary string
        boundary = "----WebKitFormBoundary" + "".join(random.choices(string.ascii_letters + string.digits, k=16))  # noqa: S311

        # Create the multipart form data manually
        body_parts = []
        for key, value in form_data.items():
            # Handle JSON fields as text
            if isinstance(value, dict):
                value = json.dumps(value)

            body_parts.append(
                f"--{boundary}\r\n" f'Content-Disposition: form-data; name="{key}"\r\n\r\n' f"{value}\r\n"
            )

        # Join all the parts with boundary and include the final boundary
        body = "".join(body_parts) + f"--{boundary}--\r\n"

        # Set headers with the custom boundary
        headers = {"Content-Type": f"multipart/form-data; boundary={boundary}"}

        def generator(_offset: int):
            # Update the offset for pagination
            form_data["variables"]["pagingInfo"]["offset"] = _offset

            # Send the POST request
            resp = requests.post(_GRAPHQL_API_URL, headers=headers, data=body)

            if not resp.ok:
                raise SilpoRequestException(f"Failed to fetch stores: {resp.text}")

            # Extract relevant data from the response
            data = resp.json()["data"]["stores"]
            return [StoreModel(**x) for x in data["items"]], data["count"]

        return Cursor(generator=generator, page_size=form_data["variables"]["pagingInfo"]["limit"])

    @classmethod
    def get_branch_id(cls, *filial_ids: int) -> list[FilialModel]:
        resp = requests.get(cls._GET_BRANCH_BY_FILIAL_ID_URL, params={"filialIds[]": filial_ids})
        resp.raise_for_status()
        return [FilialModel(**item) for item in resp.json()["items"]]


class City:
    _CITY_QUERY = """query cityWithStores($slug: String) {
          city(slug: $slug) {
            ...CityBaseFragment
            storeFilterable {
              ...StoreBaseFragment
              __typename
            }
            __typename
          }
        }

        fragment StoreBaseFragment on Store {
          id
          images {
            image {
              url
              __typename
            }
            __typename
          }
          electricityState
          isDesigned
          isLesilpo
          filial_id
          link
          title
          premium
          mapLink
          slug
          active
          cacheAmount
          terminalEnabled
          withGenerator
          withWifi
          withStarlink
          activeHours {
            start
            end
            __typename
          }
          filialType
          location {
            ...LocationBaseFragment
            __typename
          }
          city {
            ...CityBaseFragment
            __typename
          }
          __typename
        }

        fragment CityBaseFragment on City {
          id
          title
          slug
          __typename
        }

        fragment LocationBaseFragment on Location {
          lat
          lng
          __typename
        }"""

    @classmethod
    def get(cls, slug: str) -> CityModel:
        resp = requests.post(
            _GRAPHQL_API_URL,
            json={
                "query": cls._CITY_QUERY,
                "variables": {"slug": slug},
                "operationName": "cityWithStores",
            },
        )
        resp.raise_for_status()
        data = resp.json()["data"]["city"]
        return CityModel(**data) if data else None



from enum import Enum
from typing import Literal, Optional
from urllib.parse import urljoin

import requests
from cryptography.utils import cached_property
from pydantic import BaseModel, Field

from pysilpo.utils.cursor import Cursor
from pysilpo.utils.exceptions import SilpoException, SilpoRequestException


class SortBy(str, Enum):
    POPULARITY = "popularity"
    PROMOTION = "promotion"
    RATING = "rating"
    NAME = "name"
    PRICE = "price"
    PRODUCTS_LIST = "productsList"


class ProductModel(BaseModel):
    id: str = Field(..., alias="id")
    title: str = Field(..., alias="title")
    icon: str = Field(..., alias="icon")
    price: float = Field(..., alias="price")
    old_price: Optional[float] = Field(None, alias="oldPrice")
    offer_id: str = Field(..., alias="offerId")
    ratio: str = Field(..., alias="ratio")
    section_slug: str = Field(..., alias="sectionSlug")
    company_id: str = Field(..., alias="companyId")
    branch_id: str = Field(..., alias="branchId")
    external_product_id: int = Field(..., alias="externalProductId")
    promotions: list[dict] = Field(..., alias="promotions")
    special_prices: list[dict] = Field(..., alias="specialPrices")
    created_at: str = Field(..., alias="createdAt")
    slug: str = Field(..., alias="slug")
    add_to_basket_step: float = Field(..., alias="addToBasketStep")
    stock: float = Field(..., alias="stock")
    display_price: float = Field(..., alias="displayPrice")
    display_old_price: Optional[float] = Field(None, alias="displayOldPrice")
    display_ratio: str = Field(..., alias="displayRatio")
    guest_product_rating: Optional[float] = Field(None, alias="guestProductRating")
    guest_product_rating_count: Optional[int] = Field(None, alias="guestProductRatingCount")
    classifier_sap_id: Optional[str] = Field(None, alias="classifierSapId")
    origin_type: Optional[str] = Field(None, alias="originType")
    brand_id: Optional[str] = Field(None, alias="brandId")
    brand_title: Optional[str] = Field(None, alias="brandTitle")
    weighted: bool = Field(..., alias="weighted")
    blur_for_under_aged: bool = Field(..., alias="blurForUnderAged")


class CategoryModel(BaseModel):
    id: str = Field(..., alias="id")
    slug: str = Field(..., alias="slug")
    parent_id: Optional[str] = Field(None, alias="parentId")
    title: str = Field(..., alias="title")
    media: dict = Field(..., alias="media")
    tile_size: dict = Field(..., alias="tileSize")
    order: int = Field(..., alias="order")
    visibility: bool = Field(..., alias="visibility")
    updated_at: str = Field(..., alias="updatedAt")

    @cached_property
    def products(self) -> Cursor[ProductModel]:
        return Product.all(category_slug=self.slug, include_child_categories=False)


class Product:
    _DOMAIN = "https://sf-ecom-api.silpo.ua"
    _PRODUCTS_URL = urljoin(_DOMAIN, "/v1/uk/branches/{branch_id}/products")
    _CATEGORIES_URL = urljoin(_DOMAIN, "/v1/uk/branches/{branch_id}/categories")

    _DEFAULT_BRANCH_ID = "00000000-0000-0000-0000-000000000000"

    @classmethod
    def categories(cls, branch_id=_DEFAULT_BRANCH_ID) -> Cursor[CategoryModel]:
        full_url = cls._CATEGORIES_URL.format(branch_id=branch_id)

        def generator(_offset: int):
            resp = requests.get(full_url, params={"limit": 1000, "offset": _offset})
            resp.raise_for_status()
            data = resp.json()
            return [CategoryModel(**category) for category in data["items"]], data["total"]

        return Cursor[CategoryModel](generator=generator, page_size=1000)

    @classmethod
    def all(
        cls,
        branch_id=_DEFAULT_BRANCH_ID,
        category_slug: Optional[str] = None,
        search: Optional[str] = None,
        include_child_categories: bool = True,
        sort_by: SortBy = SortBy.POPULARITY,
        sort_direction: Literal["desc", "asc"] = "desc",
        delivery_type: Optional[
            Literal["DeliveryHome", "LongDelivery", "SelfPickup", "DeliveryExpressByPromise"]
        ] = None,
        in_stock: bool = False,
        limit: int = 50,
        offset: int = 0,
    ) -> Cursor[ProductModel]:
        """
        Get all products from the branch

        :param branch_id: Branch where to get products from
        :param category_slug: You can get category slug from get_categories method
        :param search: Search query
        :param include_child_categories: Do you want to include child categories of the category_slug
        :param sort_by: SortBy enum
        :param sort_direction: asc (A-Z) or desc (Z-A)
        :param delivery_type: Delivery type
        :param in_stock: Get only in stock products
        :param limit: How many products to get per request
        :param offset: How many products to skip
        :return:
        """
        # TODO: Add support for other query parameters, e.g. get data by products, productsIds, productsSlugs,
        #  category, set, mustHavePromotion, search, offersIds, isFavorite, isCarousel; Filter by etc.
        if category_slug is None and search is None:
            raise SilpoException("You must provide either category_slug or search query")
        full_url = cls._PRODUCTS_URL.format(branch_id=branch_id)
        query_params = {
            "limit": limit,
            "offset": offset,
            "category": category_slug,
            "includeChildCategories": include_child_categories,
            "sortBy": sort_by,
            "sortDirection": sort_direction,
            "inStock": in_stock,
        }
        if delivery_type:
            query_params["deliveryType"] = delivery_type
        if search:
            query_params["search"] = search

        def generator(_offset: int):
            query_params["offset"] = _offset
            resp = requests.get(full_url, params=query_params)
            if not resp.ok:
                raise SilpoRequestException(f"Failed to fetch products: {resp.text}")
            data = resp.json()
            return [ProductModel(**product) for product in data["items"]], data["total"]

        return Cursor(generator=generator, page_size=limit)

    @classmethod
    def search(
        cls,
        search: str,
        branch_id=_DEFAULT_BRANCH_ID,
        **kwargs,
    ) -> Cursor[ProductModel]:
        """
        Search for products

        :param search: Search query
        :param branch_id: Branch where to search products
        :param kwargs: Other query parameters from Product.all(...)
        :return:
        """
        return cls.all(branch_id=branch_id, search=search, sort_by=SortBy.PRODUCTS_LIST, **kwargs)



import logging
import os

from pysilpo.client import Silpo

__version__ = "1.0.2"

DEBUG = int(os.getenv("DEBUG", "0"))

logging.basicConfig(
    level=logging.DEBUG if DEBUG else logging.INFO,
    format="%(asctime)s | %(levelname)-17s | [%(name)s] | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

__all__ = ("__version__", "Silpo")



from typing import Literal, Optional

from pysilpo.services.authorization import User
from pysilpo.services.cheque import Cheque
from pysilpo.services.product import Product
from pysilpo.services.store import City, Store
from pysilpo.utils.cache import SQLiteCache
from pysilpo.utils.exceptions import SilpoAuthorizationException


class Silpo:
    product = Product
    store = Store
    city = City

    def __init__(
        self,
        phone_number: Optional[str] = None,
        otp_delivery_method: Literal["sms", "viber-sms"] = "sms",
    ):
        self._user = None
        self._cheque = None
        if phone_number is not None:
            self._user = User(phone_number=phone_number).request_otp(otp_delivery_method).login()
            self._cheque = Cheque(self._user)

    @property
    def cheque(self) -> Cheque:
        if self._cheque is None:
            raise SilpoAuthorizationException(
                "User is not authorized. " "Please provide phone number e.g. Silpo(phone_number='+380123456789')"
            )
        return self._cheque

    @classmethod
    def clear_cache(cls):
        SQLiteCache().clear()



from datetime import datetime, timezone

import pytest

from pysilpo.utils.cursor import Cursor
from pysilpo.utils.utils import subtract_months


class DummyGenerator:
    def __init__(self):
        self.values = range(23)
        self.limit = 5
        self.rounded_count = 25

    def __call__(self, _offset: int):
        return list(self.values[_offset : _offset + self.limit]), len(self.values)


@pytest.fixture
def cursor():
    generator = DummyGenerator()
    return Cursor(generator=generator, page_size=generator.limit)


class TestCursor:
    def test_create_cursor(self):
        generator = DummyGenerator()
        cursor = Cursor(generator=generator, page_size=generator.limit)
        assert cursor.pages == {}
        assert cursor.page_size == generator.limit
        assert cursor.curr == 0
        assert cursor.total_count is None
        assert cursor.rounded_count is None
        assert cursor.fetched_count == 0

    def test_len(self, cursor):
        total_count = len(cursor.generator.values)
        assert len(cursor) == total_count, "Total count is not equal to len(cursor)"
        assert cursor.total_count == total_count, "Total count is not equal to len(cursor)"
        assert cursor.rounded_count == cursor.generator.rounded_count, "Round count is not equal to len(cursor)"
        assert cursor.curr == 0

        # To know total_count we need to fetch at least one chunk
        assert cursor.fetched_count == cursor.page_size

    def test_iter(self, cursor):
        for i, val in enumerate(cursor):
            assert val == i
        assert cursor.curr == 0

        assert list(cursor) == list(cursor.generator.values)
        assert cursor.curr == 0

    def test_index(self, cursor):
        assert cursor[-1] == list(cursor.generator.values)[-1]
        assert cursor[-5] == list(cursor.generator.values)[-5]

    def test_first(self, cursor):
        total_count = len(cursor.generator.values)
        assert cursor[0] == 0
        assert cursor.curr == 0
        assert cursor.total_count == total_count
        assert cursor.rounded_count == cursor.generator.rounded_count

        # To know total_count we need to fetch at least one chunk
        assert cursor.fetched_count == cursor.page_size

    def test_slice(self, cursor):
        total_count = len(cursor.generator.values)
        assert cursor[1:3] == [1, 2]
        assert cursor.curr == 0
        assert cursor.total_count == total_count
        assert cursor.rounded_count == cursor.generator.rounded_count

        # To know total_count we need to fetch at least one chunk
        assert cursor.fetched_count == cursor.page_size

        assert cursor[1 : cursor.page_size * 2 : 2] == list(range(1, cursor.page_size * 2, 2))
        assert cursor.curr == 0
        assert cursor.total_count == total_count
        assert cursor.rounded_count == cursor.generator.rounded_count

        # To know total_count we need to fetch at least one chunk
        assert cursor.fetched_count == cursor.page_size * 2

        assert cursor[5:1:-1] == [5, 4, 3, 2]

    def test_slice_only_start(self, cursor):
        total_count = len(cursor.generator.values)
        assert cursor[1:] == list(range(1, total_count))
        assert cursor.curr == 0
        assert cursor.total_count == total_count
        assert cursor.rounded_count == cursor.generator.rounded_count
        assert cursor.fetched_count == total_count

    def test_slice_only_stop(self, cursor):
        total_count = len(cursor.generator.values)
        assert cursor[:3] == list(range(3))
        assert cursor.curr == 0
        assert cursor.total_count == total_count
        assert cursor.rounded_count == cursor.generator.rounded_count
        assert cursor.fetched_count == cursor.page_size

    def test_slice_only_step(self, cursor):
        total_count = len(cursor.generator.values)
        assert cursor[::2] == list(range(0, total_count, 2))
        assert cursor.curr == 0
        assert cursor.total_count == total_count
        assert cursor.rounded_count == cursor.generator.rounded_count
        assert cursor.fetched_count == total_count

    def test_index_error(self, cursor):
        with pytest.raises(IndexError):
            cursor[1000]

        assert cursor[1000:2000] == []

        assert cursor[1000:2000:2] == []

        assert cursor[1000:] == []

        assert cursor[:1000] == list(cursor.generator.values)

        assert cursor[1000::2] == []


class TestSubtractMonths:
    """Test suite for the subtract_months function."""

    def test_basic_subtraction(self):
        """Test basic month subtraction that doesn't cross year boundary"""
        date = datetime(2025, 3, 15)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 2, 15)

        date = datetime(2025, 5, 10)
        result = subtract_months(date, 2)
        assert result == datetime(2025, 3, 10)

    def test_cross_year_boundary(self):
        """Test subtraction that crosses year boundary"""
        date = datetime(2025, 1, 15)
        result = subtract_months(date, 1)
        assert result == datetime(2024, 12, 15)

        date = datetime(2025, 3, 10)
        result = subtract_months(date, 15)
        assert result == datetime(2023, 12, 10)

    def test_multiple_year_cross(self):
        """Test subtraction that crosses multiple years"""
        date = datetime(2025, 5, 20)
        result = subtract_months(date, 25)
        assert result == datetime(2023, 4, 20)

    def test_end_of_month_handling(self):
        """Test handling of month end dates"""
        # January 31 -> December 31 (same day)
        date = datetime(2025, 1, 31)
        result = subtract_months(date, 1)
        assert result == datetime(2024, 12, 31)

        # March 31 -> February (should adjust to Feb 28/29)
        date = datetime(2025, 3, 31)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 2, 28)

        # Leap year test
        date = datetime(2024, 3, 31)
        result = subtract_months(date, 1)
        assert result == datetime(2024, 2, 29)

    def test_month_length_differences(self):
        """Test transitions between months with different lengths"""
        # May 31 -> April 30
        date = datetime(2025, 5, 31)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 4, 30)

        # May 30 -> April 30 (same day)
        date = datetime(2025, 5, 30)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 4, 30)

        # December 31 -> November 30
        date = datetime(2025, 12, 31)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 11, 30)

    def test_preserves_time_components(self):
        """Test that time components are preserved"""
        date = datetime(2025, 3, 15, 14, 30, 45, 123456)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 2, 15, 14, 30, 45, 123456)

    def test_preserves_timezone(self):
        """Test that timezone information is preserved"""
        date = datetime(2025, 3, 15, 14, 30, 45, tzinfo=timezone.utc)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 2, 15, 14, 30, 45, tzinfo=timezone.utc)

    def test_zero_months(self):
        """Test subtracting zero months returns the same date"""
        date = datetime(2025, 3, 15)
        result = subtract_months(date, 0)
        assert result == date

    def test_negative_months(self):
        """Test that negative months raises a ValueError"""
        date = datetime(2025, 3, 15)
        with pytest.raises(ValueError):
            subtract_months(date, -1)

    def test_large_month_subtraction(self):
        """Test with large number of months"""
        date = datetime(2025, 3, 15)
        result = subtract_months(date, 100)
        assert result == datetime(2016, 11, 15)

    def test_february_edge_cases(self):
        """Test specific February edge cases, especially around leap years"""
        # Non-leap year: March 29, 30, 31 -> February 28
        date = datetime(2025, 3, 29)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 2, 28)

        date = datetime(2025, 3, 30)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 2, 28)

        date = datetime(2025, 3, 31)
        result = subtract_months(date, 1)
        assert result == datetime(2025, 2, 28)

        # Leap year: March 29 -> February 29
        date = datetime(2024, 3, 29)
        result = subtract_months(date, 1)
        assert result == datetime(2024, 2, 29)

        # Leap year: March 30, 31 -> February 29
        date = datetime(2024, 3, 30)
        result = subtract_months(date, 1)
        assert result == datetime(2024, 2, 29)

        date = datetime(2024, 3, 31)
        result = subtract_months(date, 1)
        assert result == datetime(2024, 2, 29)

    @pytest.mark.parametrize(
        "input_date,months,expected",
        [
            # Basic cases
            (datetime(2025, 5, 15), 1, datetime(2025, 4, 15)),
            (datetime(2025, 5, 15), 3, datetime(2025, 2, 15)),
            # Year boundary
            (datetime(2025, 2, 15), 3, datetime(2024, 11, 15)),
            # Month length adjustments
            (datetime(2025, 5, 31), 1, datetime(2025, 4, 30)),
            (datetime(2025, 3, 31), 1, datetime(2025, 2, 28)),
            (datetime(2024, 3, 31), 1, datetime(2024, 2, 29)),
        ],
    )
    def test_parameterized_cases(self, input_date, months, expected):
        """Test multiple scenarios with parameterization"""
        assert subtract_months(input_date, months) == expected



import math
from typing import Generic, Protocol, TypeVar, Union

T = TypeVar("T")


class Empty:
    pass


class Generator(Protocol):
    def __call__(self, _offset: int) -> tuple[list[T], int]:
        ...


class Cursor(Generic[T]):
    def __init__(self, generator: Generator, page_size: int):
        self.generator = generator
        self.total_count = None
        self.rounded_count = None
        self.fetched_count = 0
        self.pages = {}
        self.page_size = page_size
        self.curr = 0

    def fetch_new_page(self, index: int) -> list[T]:
        page_index = math.floor(index // self.page_size)
        if self.rounded_count is not None and index > self.rounded_count:
            raise IndexError
        page_content, total_count = self.generator(_offset=page_index * self.page_size)
        self.total_count = total_count

        # We need rounded count to know how many items we have in total, because we can't rely on total_count
        self.rounded_count = math.ceil(total_count / self.page_size) * self.page_size
        if not page_content:
            raise IndexError
        if page_index not in self.pages:
            self.fetched_count += len(page_content)
        self.pages[page_index] = page_content
        return self.pages[page_index]

    def get_page(self, index: int) -> list[T]:
        try:
            return self.pages[math.floor(index // self.page_size)]
        except KeyError:
            return self.fetch_new_page(index)

    def get(self, index: int) -> Union[T, Empty]:
        try:
            return self.get_page(index)[index % self.page_size]
        except IndexError:
            return Empty

    def first(self) -> T:
        return self[0]

    def __getitem__(self, index: Union[int, slice]) -> Union[list[T], T]:
        if isinstance(index, slice):
            return [
                self.get(i)
                for i in range(index.start or 0, index.stop or len(self), index.step or 1)
                if self.get(i) is not Empty
            ]
        if index < 0:
            index = len(self) + index
        val = self.get(index)
        if val is Empty:
            raise IndexError
        return val

    def __iter__(self):
        return self

    def __next__(self) -> T:
        try:
            val = self[self.curr]
        except IndexError:
            self.curr = 0
            raise StopIteration from None
        self.curr += 1
        return val

    def __len__(self) -> int:
        """
        Some API endpoints might return more items than they return in `total`
        :return: Total count of fetched items if it's more than `total` or `total` otherwise
        """
        if self.total_count is None:
            try:
                self.fetch_new_page(0)
            except IndexError:
                self.total_count = 0
                return 0
        return self.fetched_count if self.fetched_count >= self.total_count else self.total_count

    def __repr__(self):
        return f"<Cursor len={len(self)}> at {hex(id(self))}"



import logging
from datetime import datetime

import jwt


def get_logger(name: str = "PySilpo") -> logging.Logger:
    return logging.getLogger(name)


def get_jwt_expires_in(jwt_token: str) -> int:
    decoded = jwt.decode(jwt_token, options={"verify_signature": False})
    return decoded["exp"]


def subtract_months(date: datetime, months: int) -> datetime:
    """
    Subtract a specified number of months from a datetime object.

    Args:
        date: A datetime object to subtract months from
        months: Number of months to subtract (positive integer)

    Returns:
        A new datetime object with the months subtracted
    """
    if months < 0:
        raise ValueError("months should be a positive integer")

    year = date.year
    month = date.month

    month -= months

    while month <= 0:
        year -= 1
        month += 12

    if date.day > 28:
        if month == 2:
            last_day = 29 if (year % 4 == 0 and year % 100 != 0) or (year % 400 == 0) else 28
        elif month in [4, 6, 9, 11]:
            last_day = 30
        else:
            last_day = 31

        new_day = min(date.day, last_day)
    else:
        new_day = date.day

    return datetime(
        year=year,
        month=month,
        day=new_day,
        hour=date.hour,
        minute=date.minute,
        second=date.second,
        microsecond=date.microsecond,
        tzinfo=date.tzinfo,
    )
